#!/opt/alt/python37/bin/python3
# coding:utf-8

import sys
import json
import logging
import configparser
from services import RequestService, MySQLService, ExternalService
from objects import get_random_id


conf = configparser.ConfigParser()


def init_config(confname='/opt/alt/php-xray/elastic_agent.ini'):
    """
    Read and set up config file
    :param confname: path to config to read
    """
    global conf
    config_exists = conf.read(confname)
    if not config_exists:
        logging.error('Failed to read config %s', confname)
        raise RuntimeError('Failed to read config %s' % confname)


def configure_logging(logname='/tmp/agent.log'):
    """
    Configure logging: enabled or disabled
    :param logname: path to log
    :return: tuple(logging level, path to log)
    """
    turned = conf['agent']['logging']
    level_map = {
        'On': logging.DEBUG,
        'Off': logging.CRITICAL
    }
    level = level_map.get(turned, logging.CRITICAL)
    logging.basicConfig(filename=logname, level=level, format='%(asctime)s %(message)s',
                        datefmt='%m/%d/%Y %I:%M:%S %p')
    return level, logname


def get_client_data(stream=sys.stdin):
    """
    Get data from php client. JSON is expected
    :param stream: data provider
    :return: JSON-decoded data (dict)
    """
    for line in stream:
        logging.info('Input: %s', line)
        try:
            data = json.loads(line)
            logging.info('Fetched data: %s', data)
            return data
        except Exception as e:
            # may be sent to Sentry
            logging.error('Failed to parse <%s> with error: %s', line, e)
            raise


if __name__ == "__main__":
    init_config()
    configure_logging()
    client_data = get_client_data()
    # generate trace id
    current_trace = get_random_id(128)
    # get service name
    s_name = client_data['service_name']
    # create parent transaction (whole request)
    req = RequestService(s_name, current_trace)
    req(client_data)
    # create MySQL and external transactions
    for s in (MySQLService, ExternalService):
        serv = s(s_name, current_trace, parent=req.get_transaction())
        serv(client_data)
