"""
This module contains classes describing errors and different types of transactions and spans
"""
import logging
import getpass
import pwd
import random


def get_random_id(length):
    """
    Generate a random hex ID of given bits length
    :param length: number of bits in id
    :return: hex encoded random ID
    """
    return hex(random.getrandbits(length))[2:]


class DurationTuner(type):
    """
    Metaclass for tuning duration to milliseconds
    """
    def __call__(self, *args, **kwargs):
        try:
            kwargs['duration'] *= 1000
            kwargs['timestamp'] = int(''.join(str(kwargs['timestamp']).split('.')))
        except KeyError:
            pass
        return super().__call__(*args, **kwargs)


class BaseTransaction(dict, metaclass=DurationTuner):
    """
    Base transaction class.
    Extends given data with ids and additional fields and packs whole structure under transaction key.
    Also provides tuning duration to milliseconds.
    """
    def __init__(self, trace, **kwargs):
        self.transaction_data = {
            "id": get_random_id(64),
            "trace_id": trace,
            "span_count": {"started": 0, 'dropped': 0},
            'sampled': True
        }
        self.transaction_data.update(kwargs)
        super().__init__(transaction=self.transaction_data)
        logging.info('Generated transaction: %s', self)

    def __getitem__(self, item):
        return self.transaction_data[item]


class RequestTransaction(BaseTransaction):
    """
    Class describing transaction of 'request' type.
    Adds full URL field, name field, user info.
    """
    def __init__(self, trace, **kwargs):
        try:
            kwargs.update({
                'type': 'request',
                'name': self.name(kwargs['context']['request'])
            })
            kwargs['context']['request']['url']['full'] = self.full_url(kwargs['context']['request'])
        except KeyError as e:
            # may be sent to Sentry
            logging.error('Failed to create Transaction: %s is required!', e)
            raise
        super().__init__(trace, **kwargs)

    @staticmethod
    def name(request_context):
        """
        Construct transaction name from given context.
        Name includes HTTP method, host and path of the request.
        :param request_context: ['context']['request'] structure
        :return: generated name
        """
        return '{0} {1}{2}'.format(request_context['method'],
                                   request_context['url']['hostname'],
                                   request_context['url']['pathname'])

    @staticmethod
    def full_url(request_context):
        """
        Generate full URL: {protocol}://{host}/{path} from given context
        :param request_context: ['context']['request'] structure
        :return: generated URL
        """
        return '{0}://{1}{2}'.format(request_context['url']['protocol'],
                                     request_context['url']['hostname'],
                                     request_context['url']['pathname'])


class MySQLTransaction(BaseTransaction):
    """
    Class describing database fake transaction -- total MySQL time.
    Adds name field and parent id.
    """
    def __init__(self, trace, parent_transact, **kwargs):
        try:
            kwargs.update({
                'type': 'backgroundjob',
                'name': 'Total time MySQL {0}'.format(parent_transact['name']),
                'parent_id': parent_transact['id']
            })
        except KeyError as e:
            # may be sent to Sentry
            logging.error('Failed to create Transaction: %s is required!', e)
            raise
        super().__init__(trace, **kwargs)


class ExternalTransaction(BaseTransaction):
    """
    Class describing external fake transaction -- total External time.
    Adds name field and parent id.
    """
    def __init__(self, trace, parent_transact, **kwargs):
        try:
            kwargs.update({
                'type': 'outboundrequest',
                'name': 'Total time external {0}'.format(parent_transact['name']),
                'parent_id': parent_transact['id']
            })
        except KeyError as e:
            # may be sent to Sentry
            logging.error('Failed to create Transaction: %s is required!', e)
            raise
        super().__init__(trace, **kwargs)


class BaseSpan(dict, metaclass=DurationTuner):
    """
    Base span class.
    Extends given data with ids and packs whole structure under span key.
    Also provides tuning duration to milliseconds.
    """
    def __init__(self, trace, parent_transaction, **kwargs):
        self.span_data = {
            "id": get_random_id(64),
            "trace_id": trace,
            "parent_id": parent_transaction,
            "transaction_id": parent_transaction
        }
        self.span_data.update(kwargs)
        super().__init__(span=self.span_data)
        logging.info('Generated span: %s', self)


class MySQLSpan(BaseSpan):
    """
    Class describing span of 'db' type -- database span.
    """
    def __init__(self, trace, parent_transaction, **kwargs):
        kwargs.update({
            "type": "db",
            "subtype": "mysql"
        })
        super().__init__(trace, parent_transaction, **kwargs)


class ExternalSpan(BaseSpan):
    """
    Class describing span of 'external' type.
    """
    def __init__(self, trace, parent_transaction, **kwargs):
        kwargs.update({
            "type": "external",
            "subtype": "http"
        })
        super().__init__(trace, parent_transaction, **kwargs)


class Error(dict):
    """
    Wrapper class for error objects
    """
    def __init__(self, parent_transaction, trace, **kwargs):
        self.error_data = {
            "id": get_random_id(64),
            "trace_id": trace,
            "parent_id": parent_transaction,
            "transaction_id": parent_transaction
        }
        self.error_data.update(kwargs)
        super().__init__(error=self.error_data)
        logging.info('Generated error: %s', self)
