"""
This module contains classes describing defined APM services: parent (whole request), database and external
"""
import logging
import json
import configparser
import urllib.request as req
import urllib.error as err
from objects import RequestTransaction,  MySQLTransaction, ExternalTransaction, BaseSpan, MySQLSpan, ExternalSpan, Error

conf = configparser.ConfigParser()
conf.read('/opt/alt/php-xray/elastic_agent.ini')


class BaseService:
    """
    Abstract class describing basic service.
    Implements common methods for processing input JSON gained from php-xray module.
    When called, extracts corresponding data and sends it to APM server.
    """
    def __init__(self, service_name, trace_id, parent=None):
        self.trace_id = trace_id
        self.parent = parent
        self.metadata = None
        self.transaction = None
        self.spans = list()
        self.errors = list()
        self.build_metadata(self.process_serv_name(service_name))

    def __call__(self, input_data):
        self.extract(input_data)
        self.send()

    def process_serv_name(self, name):
        """
        Service name processor
        :param name: initial name
        :return: changed name
        """
        return name

    def build_metadata(self, s_name):
        """
        Construct service metadata
        """
        def validate(s):
            """
            Replace all non-allowed symbols with '_' symbol
            :param s: string to validate
            :return: fixed string
            """
            return s.replace('.', '_')

        logging.info('Received service name: %s', s_name)
        self.metadata = {
            "metadata":
                {
                    "service":
                        {
                            "name": validate(s_name),
                            "agent": {"version": "0.1", "name": conf['agent']['name']}
                        }
                }
        }
        logging.info('Built metadata %s', self.metadata)

    def set_transaction(self, data):
        """
        Method aimed to construct corresponding transaction
        :param data: input JSON from php-xray module
        :return: should return whole data structure regarding current transaction, including spans and errors
        """
        raise NotImplementedError('set_transaction method must be defined')

    def set_span(self):
        """
        Method aimed to set corresponding span
        :return: span object
        """
        return BaseSpan

    def get_transaction(self):
        """
        Get generated transaction
        :return: transaction object
        """
        return self.transaction

    @staticmethod
    def flush_transaction_data(full_data):
        """
        Clear transaction data from spans and errors (needed for clear transaction construction)
        :param full_data: whole transaction data structure including spans and errors
        :return: transaction data cleared from spans and errors
        """
        transaction_data = {key: value for key, value in full_data.items() if key not in ('spans', 'errors')}
        logging.info('Flushed transaction %s', transaction_data)
        return transaction_data

    def extract(self, input_data):
        """
        Build corresponding object: transaction, spans list, errors list
        :param input_data: input JSON data from php-xray module
        """
        raw_transaction_data = self.set_transaction(input_data)
        span = self.set_span()
        try:
            self.spans = [span(self.trace_id, self.transaction['id'], **_span) for _, _span in
                          filter(lambda x: 'span' in x[0], raw_transaction_data.get('spans', {}).items())]
            self.errors = [Error(self.trace_id, self.transaction['id'], **_error) for _, _error in
                           filter(lambda x: 'error' in x[0], raw_transaction_data.get('errors', {}).items())]
        except Exception as e:
            # may be sent to Sentry
            logging.error('Failed to attach spans and errors: %s', e)
            raise

    def body(self):
        """
        Prepare body for sending to APM server
        :return: packed service data
        """
        def pack(*args):
            """
            Pack objects for sending to APM server
            :param args: objects to pack for sending to APM server
            :return: serialized data ready to be sent to APM server
            """
            def serialize(struct):
                """
                Serialize to format accepted by APM server
                :param struct: data to serialize (dict)
                :return: JSON-encoded bytes
                """
                return (json.dumps(struct) + '\n').encode('utf-8')
            return b''.join([serialize(arg) for arg in args])

        return pack(self.metadata, self.transaction, *self.spans, *self.errors)

    def send(self):
        """
        POST given body to APM server
        :return: server response status
        """
        endpoint = '{protocol}://{server}:{port}/intake/v2/events'.format(protocol=conf['apm_server']['proto'],
                                                                          server=conf['apm_server']['host'],
                                                                          port=conf['apm_server']['port'])
        headers = {
            "Content-Type": "application/x-ndjson"
        }
        try:
            r = req.urlopen(req.Request(endpoint, data=self.body(), headers=headers))
            logging.info('Sent to APM server: %s %s', r.status, r.reason)
            return r.status
        except err.HTTPError as e:
            logging.error('Sent to APM server: %s %s %s', e.code, e.reason, e.read())
            return e.code


class RequestService(BaseService):
    """
    Class describing service for processing parent transaction (whole request)
    """
    def set_transaction(self, data):
        """
        Construct Request transaction
        :param data: input JSON from php-xray module
        :return: whole data structure, including spans and errors
        """
        full_data = data['parent_transaction']
        self.transaction = RequestTransaction(self.trace_id, **self.flush_transaction_data(full_data))
        return full_data


class MySQLService(BaseService):
    """
    Class describing service for processing MySQL transaction
    """
    def process_serv_name(self, name):
        """
        Service name processor
        :param name: initial name
        :return: initial name + database postfix
        """
        return name + '_database'

    def set_transaction(self, data):
        """
        Construct MySQL transaction
        :param data: input JSON from php-xray module
        :return: whole data structure, including spans and errors
        """
        full_data = data['mysql_transaction']
        self.transaction = MySQLTransaction(self.trace_id, self.parent, **self.flush_transaction_data(full_data))
        return full_data

    def set_span(self):
        """
        Set MySQL span
        :return: MySQL span object
        """
        return MySQLSpan


class ExternalService(BaseService):
    """
    Class describing service for processing External transaction
    """
    def process_serv_name(self, name):
        """
        Service name processor
        :param name: initial name
        :return: initial name + external postfix
        """
        return name + '_external'

    def set_transaction(self, data):
        """
        Construct External transaction
        :param data: input JSON from php-xray module
        :return: whole data structure, including spans and errors
        """
        full_data = data['external_transaction']
        self.transaction = ExternalTransaction(self.trace_id, self.parent, **self.flush_transaction_data(full_data))
        return full_data

    def set_span(self):
        """
        Set External span
        :return: External span object
        """
        return ExternalSpan
