# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

"""
This module contains contains classes implementing X-Ray Agent behaviour
"""
import json
import logging
import time
from threading import Thread, current_thread
from typing import Any

from .apiclient import get_client
from .internal.constants import agent_sock, local_tasks_storage
from .internal.exceptions import XRayError
from .internal.fault_detector import FaultDetector
from .internal.types import Task
from .internal.user_plugin_utils import extract_creds
from .internal.utils import read_sys_id, filelock, create_socket, write_sys_id, \
    dbm_storage, retry, request_id_storage, \
    get_current_cpu_throttling_time
from .manager import initialize_manager


class Agent:
    """
    X-Ray Agent class
    """

    def __init__(self):
        self.logger = logging.getLogger('agent')
        self.sys_id = read_sys_id()
        write_sys_id(self.sys_id)
        # initialize ClickHouse API client to send requests data
        clickhouse_client_object = get_client('api_req')
        self.send_client = clickhouse_client_object(system_id=self.sys_id)
        # initial state of MongoDB API client to interact with tasks
        self.task_client_object = get_client()
        # load known tasks
        self.known_tasks = self.load_tasks()
        # initialize Fault Detector
        self.fault_detector = FaultDetector()
        # start serving incoming connections
        self.listen()

    def load_tasks(self):
        """
        Load tasks from local task storage
        :return: dict in the form of {fake_id:real_task}
        """

        def _load(task_id):
            cli = self.task_client_object(system_id=self.sys_id,
                                          tracing_task_id=task_id)
            try:
                return cli.get_task()
            except XRayError as err:
                # expected Task instance returns as None!!!
                self.logger.info('Failed to load task %s with %s',
                                 task_id, str(err))

        loaded_tasks = dict()
        try:
            with dbm_storage(local_tasks_storage) as task_storage:
                for k in task_storage.keys():
                    _t = _load(task_storage[k].decode())
                    if _t is not None:
                        loaded_tasks[k.decode()] = _t
        except RuntimeError as e:
            self.logger.warning(str(e))
        return loaded_tasks

    def listen(self) -> None:
        """
        Start listening socket
        """
        _socket = create_socket(agent_sock)
        while True:
            connection, address = _socket.accept()
            t = Thread(target=self.handle, args=(connection,))
            t.start()
            self.logger.debug('[%s] Started', t.getName())

    def add_limit_faults(self, data, t_key, cpu_value):
        """
        Calculate throttling values and update given data
        with resulting throttling stat
        """
        data['hitting_limits'], data['throttled_time'] = self.fault_detector(
            t_key, cpu_value)

    def handle(self, connection: 'socket object') -> None:
        """
        Handle incoming connection
        :param connection: socket object usable to
        send and receive data on the connection
        """
        _pid, _uid, _gid = extract_creds(connection)
        current_cpu = get_current_cpu_throttling_time(_uid)
        fileobj = connection.makefile(errors='ignore')
        try:
            input_data = self.read_input(fileobj)
        except json.JSONDecodeError as e:
            self.logger.error('JSON decode failed: %s',
                              str(e), extra={'t_name': current_thread().name})
            connection.close()
            return

        if input_data is None:
            # save current CPU throttling time and timestamp
            # self.fault_detector.save(_pid, current_cpu)
            self.fault_detector.save(_pid, current_cpu)
            # attempt to flush expired entries
            self.fault_detector.flush()
            connection.close()
            return

        # otherwise calculate throttling fact, add it to data
        # and send gathered stat to CH
        try:
            if input_data.get('hitting_limits') is None:
                # only calculate faults if extension failed to get them itself
                self.add_limit_faults(input_data, _pid, current_cpu)
            self.__call__(input_data)
        except XRayError:
            pass
        finally:
            connection.close()

    def read_input(self, fileio: 'file object') -> Any:
        """
        Read input data and return decoded json
        :param fileio: a file-like object providing read method
        """
        data = fileio.read()
        self.logger.info('[%s] I received %i bytes trace',
                         current_thread().getName(), len(data.encode()))
        self.logger.debug('[%s] I received data: %s',
                          current_thread().getName(), data)
        if len(data.strip()) == 0:
            return
        return json.loads(data.strip(), strict=False)

    def __call__(self, decoded_data: dict) -> None:
        """
        Process received data
        """
        self.logger.info('[%s] Processing trace for task %s (%s)',
                         current_thread().getName(),
                         decoded_data.get('tracing_task_id'),
                         decoded_data.get('url'))
        apiclient = self.instantiate_mongo_client(
            self.get_task_id(decoded_data))
        self.process_request_data(decoded_data, apiclient)

    @retry
    def instantiate_mongo_client(self,
                                 fake_task_id: str) -> 'APIClient instance':
        """
        Initialize MongoDB client for current task
        """
        real_id = self.known_tasks[fake_task_id].task_id
        return self.task_client_object(system_id=self.sys_id,
                                       tracing_task_id=real_id)

    @staticmethod
    def get_task_id(data: dict) -> str:
        """
        Retrieve tracing task ID from the input data
        :return: tracing task ID (fake one)
        """
        return data['tracing_task_id']

    def process_request_data(self, request_data: dict, apiclient: 'APIClient instance') -> None:
        """
        Increment request ID in /usr/share/alt-php-xray/requests/{tracing_task_id} file
        Substitute request_id and tracing_task_id in request_data.
        Send request_data to ClickHouse
        :param request_data: original request data
        :param apiclient: MongoDB client instance
        """
        t = current_thread().getName()
        with request_id_storage(self.get_task_id(request_data)) as req_id_file:
            with filelock(req_id_file):
                task = apiclient.get_task()
                if task.tracing_count <= 0:
                    self.logger.info('Tracing count is 0, nothing should be done')
                    return
                # read stored request_id
                tmp_id = int(req_id_file.read().strip())
                self.logger.info('[%s] Read ID %i', t, tmp_id)
                task.request_count = tmp_id
                # update input data with stored request_id
                updated_request_data = self.update_request_data(request_data,
                                                                task)
                # send data with updated ids
                self.logger.info('[%s] Sending to ClickHouse', t)
                self.send_client(updated_request_data)
                # then increment request_id counter
                new_id = tmp_id + 1
                req_id_file.seek(0)
                req_id_file.write(str(new_id))
                self.logger.info('[%s] Incremented request ID to %i', t, new_id)
                self.update_counts(apiclient, task)

    def update_request_data(self, data: dict, task: Task) -> dict:
        """
        Substitute request_id and tracing_task_id
        :param data: original input
        :param task: a Task instance
        :return: updated input
        """
        data['request_id'] = task.request_count
        data['tracing_task_id'] = task.task_id
        for item in data['data']:
            item['request_id'] = task.request_count
            item['tracing_task_id'] = task.task_id
        self.logger.info('[%s] Input updated => %s & %s',
                         current_thread().getName(),
                         data.get('tracing_task_id'), data.get('request_id'))
        self.logger.debug('[%s] Full updated input %s',
                          current_thread().getName(), data)
        return data

    def update_counts(self, client: 'APIClient instance', task: Task) -> None:
        """
        Decrement tracing_count for tasks with tracing_by==request_qty
        :param client: MongoDB API client instance
        :param task: tracing task instance
        """
        t = current_thread().getName()
        if task.tracing_by == 'time':
            client.update_counts_only(request_count=task.request_count)
        else:
            self.logger.info('[%s] Processing request_qty', t)
            task.tracing_count -= 1
            self.logger.info('[%s] %i requests left', t, task.tracing_count)
            client.update_counts_only(tracing_count=task.tracing_count,
                                      request_count=task.request_count)
            if task.tracing_count <= 0:
                self.complete_task(task)

    def complete_task(self, _task: Task) -> None:
        """
        Stop and complete request_qty task
        :param _task: tracing task to stop
        """
        self.logger.info('[%s] Task %s should be completed',
                         current_thread().getName(),
                         _task.task_id)
        # delay for MongoDB to process counts, received lately (see XRAY-87)
        time.sleep(1)
        manager_instance = initialize_manager(self.sys_id)
        manager_instance.stop(_task.task_id)
