# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

"""
This module contains an argparse command line parser
for cloudlinux-xray-manager
"""
import argparse
import logging


def cmd_parser_manager() -> 'argparse.ArgumentParser':
    """
    Command line parser for X Ray Manager using built-in argparse module
    :return: parser
    """
    parser = argparse.ArgumentParser(prog='cloudlinux-xray-manager',
                                     description='Utility to manage X-Ray Client')
    subparsers = parser.add_subparsers(title='Commands', dest='command',
                                       required=True)

    start_parser = subparsers.add_parser('start',
                                         help='Start monitoring of URL by IP for COUNT of minutes or requests',
                                         argument_default=argparse.SUPPRESS)
    start_parser.add_argument('--system_id', metavar='SYS_ID', required=True,
                              help='unique ID of CL+ system')
    start_parser.add_argument('--url', metavar='URL', required=True,
                              help='monitoring URL')
    start_parser.add_argument('--client_ip', metavar='IP', default='*',
                              help='client IP to monitor requests from')
    counters = start_parser.add_mutually_exclusive_group()
    counters.add_argument('--time', metavar='COUNT', type=int,
                          help='number of minutes to monitor')
    counters.add_argument('--request_qty', metavar='COUNT', type=int,
                          help='number of requests to monitor')

    stop_parser = subparsers.add_parser('stop',
                                        help='Stop monitoring of given TASK')
    stop_parser.add_argument('--system_id', metavar='SYS_ID', required=True,
                             help='unique ID of CL+ system')
    stop_parser.add_argument('--tracing_task_id', metavar='TASK',
                             required=True, help='unique ID of tracing task')

    continue_parser = subparsers.add_parser('continue',
                                            help='Continue monitoring of given TASK')
    continue_parser.add_argument('--system_id', metavar='SYS_ID',
                                 required=True,
                                 help='unique ID of CL+ system')
    continue_parser.add_argument('--tracing_task_id', metavar='TASK',
                                 required=True,
                                 help='unique ID of tracing task')

    complete_parser = subparsers.add_parser('complete',
                                            help='Complete given TASK')
    complete_parser.add_argument('--system_id', metavar='SYS_ID',
                                 required=True,
                                 help='unique ID of CL+ system')
    complete_parser.add_argument('--tracing_task_id', metavar='TASK',
                                 required=True,
                                 help='unique ID of tracing task')

    delete_parser = subparsers.add_parser('delete',
                                          help='Delete given TASK')
    delete_parser.add_argument('--system_id', metavar='SYS_ID',
                               required=True,
                               help='unique ID of CL+ system')
    delete_parser.add_argument('--tracing_task_id', metavar='TASK',
                               required=True,
                               help='unique ID of tracing task')

    enable_parser = subparsers.add_parser('enable-continuous',
                                          help='Schedule a URL for continuous monitoring')
    enable_parser.add_argument('--system_id', metavar='SYS_ID',
                               required=True,
                               help='unique ID of CL+ system')
    enable_parser.add_argument('--url', metavar='URL', required=True,
                               help='monitoring URL')
    enable_parser.add_argument('--email', metavar='EMAIL', required=True,
                               help='e-mail address for reporting')

    disable_parser = subparsers.add_parser('disable-continuous',
                                           help='Schedule to stop continuous monitoring for URL')
    disable_parser.add_argument('--system_id', metavar='SYS_ID',
                                required=True,
                                help='unique ID of CL+ system')
    disable_parser.add_argument('--url', metavar='URL', required=True,
                                help='monitoring URL')

    con_start_parser = subparsers.add_parser('start-continuous',
                                             help='Include URL into continuous monitoring')
    con_start_parser.add_argument('--system_id', metavar='SYS_ID',
                                  required=True,
                                  help='unique ID of CL+ system')
    con_start_parser.add_argument('--url', metavar='URL', required=True,
                                  help='monitoring URL')

    con_stop_parser = subparsers.add_parser('stop-continuous',
                                            help='Exclude URL from continuous monitoring')
    con_stop_parser.add_argument('--system_id', metavar='SYS_ID',
                                 required=True,
                                 help='unique ID of CL+ system')
    con_stop_parser.add_argument('--url', metavar='URL', required=True,
                                 help='monitoring URL')

    list_parser = subparsers.add_parser('continuous-tracing-list',
                                        help='Get list of scheduled continuous monitoring tasks')
    list_parser.add_argument('--system_id', metavar='SYS_ID',
                             required=True,
                             help='unique ID of CL+ system')

    task_list_parser = subparsers.add_parser('tasks-list',
                                             help='Get list of tasks')
    task_list_parser.add_argument('--system_id', metavar='SYS_ID',
                                  required=True,
                                  help='unique ID of CL+ system')

    req_parser = subparsers.add_parser('requests-list',
                                       help='Get list of requests for given task')
    req_parser.add_argument('--system_id', metavar='SYS_ID',
                            required=True,
                            help='unique ID of CL+ system')
    req_parser.add_argument('--tracing_task_id', metavar='TASK',
                            required=True,
                            help='unique ID of tracing task')

    req_data_parser = subparsers.add_parser('request-data',
                                            help='Get statistics for given request')
    req_data_parser.add_argument('--system_id', metavar='SYS_ID',
                                 required=True,
                                 help='unique ID of CL+ system')
    req_data_parser.add_argument('--tracing_task_id', metavar='TASK',
                                 required=True,
                                 help='unique ID of tracing task')
    req_data_parser.add_argument('--request_id', metavar='REQUEST',
                                 required=True, type=int,
                                 help='unique ID of request')

    enable_user_agent_parser = subparsers.add_parser('enable-user-agent',
                                                     help='Enable X-Ray User Agent')
    enable_user_agent_parser.add_argument('--system_id', metavar='SYS_ID',
                                          required=True,
                                          help='unique ID of CL+ system')

    disable_user_agent_parser = subparsers.add_parser('disable-user-agent',
                                                      help='Disable X-Ray User Agent')
    disable_user_agent_parser.add_argument('--system_id', metavar='SYS_ID',
                                           required=True,
                                           help='unique ID of CL+ system')

    status_user_agent_parser = subparsers.add_parser('user-agent-status',
                                                     help='Get status of X-Ray User Agent')
    status_user_agent_parser.add_argument('--system_id', metavar='SYS_ID',
                                          required=True,
                                          help='unique ID of CL+ system')

    return parser


def cmd_parser_user_manager() -> 'argparse.ArgumentParser':
    """
    Command line parser for X Ray Manager for users using built-in argparse module
    :return: parser
    """
    parser = argparse.ArgumentParser(prog='cloudlinux-xray-user-manager.py',
                                     description='Utility to manage X-Ray Client for end-users')
    subparsers = parser.add_subparsers(title='Commands', dest='command',
                                       required=True)

    start_parser = subparsers.add_parser('start',
                                         help='Start monitoring of URL by IP for COUNT of minutes or requests',
                                         argument_default=argparse.SUPPRESS)
    start_parser.add_argument('--url', metavar='URL', required=True,
                              help='monitoring URL')
    start_parser.add_argument('--client_ip', metavar='IP', default='*',
                              help='client IP to monitor requests from')
    counters = start_parser.add_mutually_exclusive_group()
    counters.add_argument('--time', metavar='COUNT', type=int,
                          help='number of minutes to monitor')
    counters.add_argument('--request_qty', metavar='COUNT', type=int,
                          help='number of requests to monitor')

    stop_parser = subparsers.add_parser('stop',
                                        help='Stop monitoring of given TASK')
    stop_parser.add_argument('--tracing_task_id', metavar='TASK',
                             required=True, help='unique ID of tracing task')

    continue_parser = subparsers.add_parser('continue',
                                            help='Continue monitoring of given TASK')
    continue_parser.add_argument('--tracing_task_id', metavar='TASK',
                                 required=True,
                                 help='unique ID of tracing task')

    complete_parser = subparsers.add_parser('complete',
                                            help='Complete given TASK')
    complete_parser.add_argument('--tracing_task_id', metavar='TASK',
                                 required=True,
                                 help='unique ID of tracing task')

    delete_parser = subparsers.add_parser('delete',
                                          help='Delete given TASK')
    delete_parser.add_argument('--tracing_task_id', metavar='TASK',
                               required=True,
                               help='unique ID of tracing task')

    subparsers.add_parser('tasks-list',
                          help='Get list of tasks')

    req_parser = subparsers.add_parser('requests-list',
                                       help='Get list of requests for given task')
    req_parser.add_argument('--tracing_task_id', metavar='TASK',
                            required=True,
                            help='unique ID of tracing task')

    req_data_parser = subparsers.add_parser('request-data',
                                            help='Get statistics for given request')
    req_data_parser.add_argument('--tracing_task_id', metavar='TASK',
                                 required=True,
                                 help='unique ID of tracing task')
    req_data_parser.add_argument('--request_id', metavar='REQUEST',
                                 required=True, type=int,
                                 help='unique ID of request')

    subparsers.add_parser('user-agent-status',
                          help='Get status of X-Ray User Agent')

    return parser


def parse_cmd_arguments(
        parser: 'argparse.ArgumentParser') -> 'argparse.Namespace':
    """
    Parse arguments with given parser
    :param parser: an instance of ArgumentParser
    :return: arguments Namespace
    """
    try:
        args = parser.parse_args()
        return args
    except SystemExit as e:
        logging.error('Invalid utility invocation', extra={'err': str(e)})
        raise
