# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

import logging
import os
import subprocess
from glob import iglob

from xray.internal.constants import request_data_storage, local_tasks_storage
from xray.internal.utils import dbm_storage

logger = logging.getLogger('reconfiguration')


def get_locations() -> tuple:
    """
    Return all possible locations of xray.ini files in glob-like form
    Panel independent
    """
    return (
        '/opt/alt/php*/link/conf',
        '/var/cagefs/*/*/etc/cl.php.d/alt-php*',
        '/opt/cpanel/ea-php*/root/etc/php.d',
        '/opt/plesk/php/*/etc/php.d',
        '/usr/local/php*/lib/php.conf.d',
        '/usr/share/cagefs/.cpanel.multiphp/opt/cpanel/ea-php*/root/etc/php.d',
        '/usr/share/cagefs-skeleton/usr/local/php*/lib/php.conf.d'
    )


def find_xray_inis(globpath: str) -> str:
    """
    Find xray.ini files in every directory matching given glob-like path.
    Return full path to xray.ini
    """
    for conf_dir in iglob(globpath):
        for name in os.listdir(conf_dir):
            if 'xray.ini' in name:
                yield os.path.join(conf_dir, name)


def remove_xray_inis() -> None:
    """
    Remove all xray.ini files found in filesystem
    """
    for location in get_locations():
        for xray_ini in find_xray_inis(location):
            logger.info('Removing %s', xray_ini)
            os.unlink(xray_ini)


def clear_dbm_file() -> None:
    """
    Remove all entries from local dbm file with fake_ids mapping
    """
    with dbm_storage(local_tasks_storage) as task_storage:
        for item in task_storage.keys():
            logger.info('Clear task %s', item.decode())
            del task_storage[item]


def remove_req_id_files() -> None:
    """
    Remove all request_id files
    """
    for name in os.listdir(request_data_storage):
        logger.info('Clear request_id for %s', name)
        os.unlink(os.path.join(request_data_storage, name))


def restart_agent() -> None:
    """
    Restart X Ray Agent
    """
    try:
        subprocess.run(['/sbin/service',
                        'xray-agent',
                        'restart'],
                       capture_output=True, text=True)
        logger.info('X Ray Agent restarted')
    except (OSError, ValueError, subprocess.SubprocessError) as e:
        logger.error('Failed to restart X-Ray Agent',
                     extra={'err': str(e)})


def reconfigure():
    """
    Perform reconfiguration
    """
    logger.info('system_id shift: reconfigure...')
    remove_xray_inis()
    remove_req_id_files()
    clear_dbm_file()
    restart_agent()
    logger.info('reconfiguration for system_id shift case completed')
