# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

"""
This module contains X-Ray User Manager service implementation
"""
import json
import logging
import re
import subprocess
from threading import Thread, current_thread
from typing import Tuple

from xray.cli_utils.validations import validate_user
from xray.internal.constants import user_agent_sock, user_agent_log
from xray.internal.exceptions import XRayError
from xray.internal.user_plugin_utils import unpack_request, pack_response, \
    extract_creds, check_for_root, error_response
from xray.internal.utils import create_socket, read_sys_id, configure_logging

logger = logging.getLogger('user_agent')


def general_exec_error() -> Tuple[bytes, bytes]:
    """
    General format of message in case of errors during manager execution
    """
    _err = 'X-Ray User Plugin failed to execute your request. Please, contact your server administrator'
    return error_response(_err).encode(), b''


def log_truncate(orig_msg: bytes) -> str:
    """
    Cut data field from the original message, because it could be huge
    """
    return re.sub('(?<="data": {).+(?=}, "result")', '...', orig_msg.decode())


def duplicate_warning_cast(orig_msg: bytes) -> bytes:
    """
    Extend warning 'Task is duplicated by URL' for end user
    """

    def gen(m):
        """
        Add more text for duplicate warning, leave others unchanged
        """
        additional = b""". \
In case if you do not see running task for the same URL in your list of tasks below, \
contact your server administrator and ask him to check whether the requested URL \
is in the tasks list of X-Ray Admin Plugin or is scheduled for continuous tracing."""
        warn = m.group(0)
        if warn == b'Task is duplicated by URL':
            return warn + additional
        return warn

    return re.sub(b'(?<="warning": ").+(?="})', gen, orig_msg)


def execute_manager(command: dict, user: str) -> Tuple[bytes, bytes]:
    """
    Trigger cloudlinux-xray-manager with requested parameters
    """
    try:
        command['system_id'] = read_sys_id()
    except XRayError:
        return general_exec_error()

    cmd = ['/usr/sbin/cloudlinux-xray-manager', command.pop('command')]
    cmd.extend([f'--{k}={v}' for k, v in command.items()])
    try:
        p = subprocess.run(cmd, capture_output=True,
                           env={'XRAYEXEC_UID': user})
    except (OSError, ValueError, subprocess.SubprocessError):
        return general_exec_error()

    byte_out, byte_err = p.stdout.strip(), p.stderr.strip()
    if byte_out:
        logger.info('[%s] Proxied command stdout: %s', current_thread().name,
                    log_truncate(byte_out))
    if byte_err:
        logger.info('[%s] Proxied command stderr: %s', current_thread().name,
                    byte_err.decode())
    return duplicate_warning_cast(byte_out), byte_err


def handle(connection: 'socket object') -> None:
    """
    Handle incoming connection
    :param connection: socket object usable to
    send and receive data on the connection
    """
    root_error = json.dumps({
        'result': 'Commands from root are not accepted'
    })
    with connection:
        _pid, _uid, _gid = extract_creds(connection)
        if check_for_root(_uid):
            connection.sendall(pack_response(root_error.encode()))
            return

        data = connection.recv(4096)
        if not data:
            return

        args = unpack_request(data)
        try:
            validate_user(args)
        except SystemExit as e:
            connection.sendall(pack_response(str(e).encode()))
            return

        _out, _err = execute_manager(args, str(_uid))
        connection.sendall(pack_response(_out or _err))


def run() -> None:
    """
    Run listening service
    """
    configure_logging(user_agent_log)
    with create_socket(user_agent_sock) as s:
        while True:
            conn, _ = s.accept()
            t = Thread(target=handle, args=(conn,))
            t.start()
            logger.info('[%s] Started', t.name)
